use crate::*;
use reqwest::{Client, Method, Request, StatusCode, Url, header::HeaderValue};
use std::{borrow::Borrow, fmt};

#[derive(Serialize)]
struct _RecordQuery<T>
where
    T: Serialize,
{
    records: Vec<T>,
}

pub struct HetznerDNSAPIClient {
    token: String,
    host: Url,
    client: Client,
}

#[derive(Debug)]
pub enum HetznerDNSAPIError {
    Unauthorized,
    Forbidden,
    NotFound,
    NotAcceptable,
    Conflict,
    UnprocessableEntity,
}

impl fmt::Display for HetznerDNSAPIError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::Unauthorized => "Missing authorization, did you set a token?",
                Self::Forbidden => "Forbidden action!",
                Self::NotFound => "Entity not found!",
                Self::NotAcceptable => "Request failed, possibly bad input!",
                Self::Conflict => "Conflict encountered",
                Self::UnprocessableEntity =>
                    "Request not procesasble, input is invalid or might already exist!",
            }
        )
    }
}
impl TryFrom<StatusCode> for HetznerDNSAPIError {
    type Error = String;
    fn try_from(value: StatusCode) -> Result<Self, Self::Error> {
        Ok(match value {
            StatusCode::UNAUTHORIZED => Self::Unauthorized,
            StatusCode::FORBIDDEN => Self::Forbidden,
            StatusCode::NOT_FOUND => Self::NotFound,
            StatusCode::NOT_ACCEPTABLE => Self::NotAcceptable,
            StatusCode::CONFLICT => Self::Conflict,
            StatusCode::UNPROCESSABLE_ENTITY => Self::UnprocessableEntity,
            _ => return Err(format!("Invalid error code: {}", value)),
        })
    }
}

impl HetznerDNSAPIClient {
    pub fn new(token: String) -> Self {
        HetznerDNSAPIClient {
            token,
            host: Url::parse("https://dns.hetzner.com/api/v1/").unwrap(),
            client: Client::new(),
        }
    }

    async fn api_call<'a, T, U, I, K, V>(
        &self,
        url: &'a str,
        method: Method,
        query: Option<I>,
        payload: Option<U>,
    ) -> Result<T, HetznerDNSAPIError>
    where
        T: for<'de> Deserialize<'de>,
        U: Serialize,
        I: IntoIterator,
        <I as IntoIterator>::Item: Borrow<(K, V)>,
        K: AsRef<str>,
        V: AsRef<str>,
    {
        let mut req = Request::new(
            method,
            self.host.join(url).map_err(|e| {
                println!("url formatting error: {}", e);
                HetznerDNSAPIError::UnprocessableEntity
            })?,
        );
        req.headers_mut().append(
            "Auth-API-Token",
            HeaderValue::from_str(self.token.as_str()).unwrap(),
        );
        if let Some(payload) = payload {
            *req.body_mut() = Some(
                serde_json::to_string(&payload)
                    .map_err(|e| {
                        println!("body encoding error: {}", e);
                        HetznerDNSAPIError::UnprocessableEntity
                    })?
                    .into(),
            );
            //println!("{}", serde_json::to_string_pretty(&payload).unwrap());
        }
        if let Some(query) = query {
            req.url_mut()
                .query_pairs_mut()
                .extend_pairs(query.into_iter());
        }
        let t = self
            .client
            .execute(req)
            .await
            .map_err(|e| {
                println!("request execution error: {}", e);
                HetznerDNSAPIError::UnprocessableEntity
            })?
            .error_for_status()
            .map_err(|e| HetznerDNSAPIError::try_from(e.status().unwrap()).unwrap())?
            .text()
            .await
            .map_err(|e| {
                println!("request decoding error: {}", e);
                HetznerDNSAPIError::UnprocessableEntity
            })?;
        serde_json::from_str::<T>(&t).map_err(|e| {
            println!("json response parsing error: {}", e);
            HetznerDNSAPIError::UnprocessableEntity
        })
    }

    pub async fn get_zones<'a>(
        &self,
        name: Option<&'a str>,
        page: Option<u32>,
        per_page: Option<u32>,
        search_name: Option<&'a str>,
    ) -> Result<Vec<Zone>, HetznerDNSAPIError> {
        let result: ZoneResult = self
            .api_call(
                "zones",
                Method::GET,
                Some(&[
                    ("name", name.unwrap_or_default()),
                    ("page", page.unwrap_or(1).to_string().as_str()),
                    ("per_page", per_page.unwrap_or(100).to_string().as_str()),
                    ("search_name", search_name.unwrap_or_default()),
                ]),
                None::<String>,
            )
            .await?;
        Ok(result.zones)
    }

    pub async fn create_zone(
        &self,
        name: String,
        ttl: Option<u64>,
    ) -> Result<Zone, HetznerDNSAPIError> {
        self.api_call(
            "zones",
            Method::POST,
            None::<&[(&str, &str); 0]>,
            Some(&[("name", name), ("ttl", ttl.unwrap_or(86400).to_string())]),
        )
        .await
    }

    pub async fn get_zone(&self, id: String) -> Result<Zone, HetznerDNSAPIError> {
        self.api_call(
            format!("zones/{}", id).as_str(),
            Method::GET,
            None::<[(&str, &str); 0]>,
            None::<&str>,
        )
        .await
    }

    pub async fn update_zone(
        &self,
        id: String,
        name: String,
        ttl: Option<u64>,
    ) -> Result<Zone, HetznerDNSAPIError> {
        self.api_call(
            format!("zones/{}", id).as_str(),
            Method::PUT,
            None::<[(&str, &str); 0]>,
            Some(&[("name", name), ("ttl", ttl.unwrap_or(86400).to_string())]),
        )
        .await
    }

    pub async fn delete_zone(&self, id: String) -> Result<(), HetznerDNSAPIError> {
        let result: () = self
            .api_call(
                format!("zones/{}", id).as_str(),
                Method::DELETE,
                None::<&[(&str, &str); 0]>,
                None::<&str>,
            )
            .await?;
        Ok(result)
    }

    pub async fn import_zone() {
        todo!()
    }
    pub async fn export_zone() {
        todo!()
    }
    pub async fn validate_zone() {
        todo!()
    }

    pub async fn get_records(
        &self,
        page: Option<u32>,
        per_page: Option<u32>,
        zone_id: Option<String>,
    ) -> Result<Vec<Record>, HetznerDNSAPIError> {
        let result: RecordsResult = self
            .api_call(
                "records",
                Method::GET,
                Some(&[
                    ("page", page.unwrap_or(1).to_string()),
                    ("per_page", per_page.unwrap_or(100).to_string()),
                    ("zone_id", zone_id.unwrap_or_default()),
                ]),
                None::<u8>,
            )
            .await?;
        Ok(result.records)
    }

    pub async fn create_record(
        &self,
        payload: RecordPayload,
    ) -> Result<Record, HetznerDNSAPIError> {
        let result: RecordResult = self
            .api_call(
                "records",
                Method::POST,
                None::<[(&str, &str); 0]>,
                Some(payload),
            )
            .await?;
        Ok(result.record)
    }

    pub async fn get_record(&self, record_id: String) -> Result<Record, HetznerDNSAPIError> {
        let result: RecordResult = self
            .api_call(
                format!("records/{}", record_id).as_str(),
                Method::GET,
                None::<[(&str, &str); 0]>,
                None::<u8>,
            )
            .await?;
        Ok(result.record)
    }

    pub async fn update_record(
        &self,
        record_id: String,
        payload: RecordPayload,
    ) -> Result<Record, HetznerDNSAPIError> {
        let result: RecordResult = self
            .api_call(
                format!("records/{}", record_id).as_str(),
                Method::PUT,
                None::<[(&str, &str); 0]>,
                Some(payload),
            )
            .await?;
        Ok(result.record)
    }

    pub async fn delete_record(&self, record_id: String) -> Result<(), HetznerDNSAPIError> {
        let result: () = self
            .api_call(
                format!("records/{}", record_id).as_str(),
                Method::DELETE,
                None::<[(&str, &str); 0]>,
                None::<u8>,
            )
            .await?;
        Ok(result)
    }

    pub async fn create_records(
        &self,
        payloads: Vec<RecordPayload>,
    ) -> Result<Vec<Record>, HetznerDNSAPIError> {
        let result: RecordsResult = self
            .api_call(
                "records/bulk",
                Method::POST,
                None::<[(&str, &str); 0]>,
                Some(_RecordQuery { records: payloads }),
            )
            .await?;
        Ok(result.records)
    }

    pub async fn update_records(
        &self,
        payloads: Vec<RecordPayload>,
    ) -> Result<Vec<Record>, HetznerDNSAPIError> {
        let result: RecordsResult = self
            .api_call(
                "records/bulk",
                Method::PUT,
                None::<[(&str, &str); 0]>,
                Some(_RecordQuery { records: payloads }),
            )
            .await?;
        Ok(result.records)
    }
}
